/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 ForTISS GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.action;

import java.util.Collection;

import org.eclipse.emf.common.command.Command;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.gef.EditDomain;
import org.eclipse.gef.commands.CommandStack;
import org.eclipse.jface.action.Action;
import org.eclipse.jface.resource.ImageDescriptor;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.services.IPersistencyService;
import org.fortiss.tooling.kernel.ui.util.EObjectSelectionUtils;

/**
 * GEF editor base implementation.
 * 
 * It provides {@link EditDomain} and thus a {@link CommandStack}.
 * 
 * @author hummel
 * @author hoelzl
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 63FF468A2D042586A22495828A0798EF
 */
public abstract class ModelElementsActionBase extends Action {

	/** Stores whether the action will modify the model. */
	private final boolean modifiesModel;

	/** Constructor. */
	protected ModelElementsActionBase(String text, String id,
			ImageDescriptor image, ImageDescriptor disabledImage,
			boolean modifiesModel) {
		super(text, image);
		setId(id);
		setDisabledImageDescriptor(disabledImage);
		this.modifiesModel = modifiesModel;
	}

	/** Returns the Command to be started in the run method. */
	protected abstract Command createCommand(Collection<EObject> selection);

	/** Returns the selection to be used for this action. */
	protected Collection<EObject> getSelection() {
		return EObjectSelectionUtils.getCurrentSelectionEObjects();
	}

	/**
	 * This method checks if the current selection is non-empty and all elements
	 * share the same (non-null) context, i.e. are from the same model.
	 */
	protected boolean isConsistentSelection() {
		ITopLevelElement context = null;
		for (EObject o : getSelection()) {
			ITopLevelElement elementContext = IPersistencyService.INSTANCE
					.getTopLevelElementContextFor(o);
			if (context == null) {
				context = elementContext;
			}
			if (elementContext == null || elementContext != context) {
				return false;
			}
		}
		return context != null;
	}

	/** {@inheritDoc} */
	@Override
	public final void run() {
		Collection<EObject> selection = getSelection();
		if (selection.isEmpty()) {
			return;
		}

		final Command cmd = createCommand(selection);
		if (cmd.canExecute()) {
			if (modifiesModel) {
				ITopLevelElement context = IPersistencyService.INSTANCE
						.getTopLevelElementContextFor(selection.iterator()
								.next());
				context.runAsCommand(new Runnable() {

					@Override
					public void run() {
						cmd.execute();
					}
				});
			} else {
				cmd.execute();
			}
		}
	}

	/** Updates the enablement of this action. */
	public void update() {
		if (!isConsistentSelection()) {
			setEnabled(false);
			return;
		}

		setEnabled(createCommand(getSelection()).canExecute());
	}
}